/*

Launcher stub for tlshell, a tlmgr gui, with hidden console window
and with UAC prompt, written in 2018 and 2020 by Siep Kroonenberg.

Based on tlmgr-gui.c for the original Perl/Tk gui from 2014/2016 and
code from tlauncher from 2016, both by Siep Kroonenberg.

We assume Windows Vista or later since TL won't even install on
earlier versions.

The relative paths of tlshell.tcl and wish or tclkit are hard-coded.
On windows, asking kpsewhich about tlshell.tcl is more trouble than
it is worth.

Public Domain
*/

#undef WINVER
#define WINVER 601
#undef _WIN32_WINNT
#define _WIN32_WINNT _WIN32_WINNT_WIN7
#undef UNICODE
#define UNICODE
#undef _UNICODE
#define _UNICODE

#include <windows.h>
#include <winnt.h>
#include <shellapi.h>
#include <shlobj.h>
#include <wtypes.h>
#include <wchar.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

static wchar_t * instroot, * texbin;
static char release[5];

#define IS_WHITESPACE(c) ((c == L' ') || (c == L'\t'))

#define FILE_EXISTS(f) ( GetFileAttributes(f) != INVALID_FILE_ATTRIBUTES )

#define DIEFLAGS MB_ICONERROR | MB_TASKMODAL | MB_SETFOREGROUND

void die( wchar_t * s, int i ) {
  MessageBox( NULL, s, L"ERROR!", DIEFLAGS );
  // PostQuitMessage( i );
  // no message loop in sight, need an explicit exit
  ExitProcess( i );
}

// free and set to NULL
#define FREE0(p) { if( p ) { free( p ); p = NULL; }}

void * calloc_fatal( size_t nitems, size_t item_size ) {
  void * p;
  p = calloc( nitems, item_size );
  if ( !p ) ExitProcess( 1 );
  return p;
}

// wide-char [v]asprintf with automatic sizing and allocation
// adapted from source code of asprintf and vasprintf
int tl_vwasprintf( wchar_t ** ret, const wchar_t * fmt, va_list ap ) {
  FILE * fp;
  wchar_t * buf;
  int len;

  fp = fopen( "NUL", "w" );
  if ( ! fp ) {
    *ret = NULL;
    return -1;
  }
  len = vfwprintf( fp, fmt, ap );
  fclose( fp );
  if( len < 0 ) return -1;
  buf = calloc_fatal( len + 1, sizeof( wchar_t ));
  if ( vswprintf( buf, len, fmt, ap ) != len ) {
    FREE0( buf );
    *ret = NULL;
    return -1;
  }
  *ret = buf;
  return len;
}

void tl_wasprintf( wchar_t ** ret, const wchar_t * fmt, ... ) {
  int len;
  va_list ap;
  va_start( ap, fmt );
  len = tl_vwasprintf( ret, fmt, ap );
  va_end( ap );
  if ( len < 0 ) die( L"tl_wasprintf: Failure to construct string", 1 );
}

void get_tl_release() {
  // read release from TL database
  wchar_t * dbfilename;
  char * cptr;
  int i, j, l;
  FILE * dbfile;
  #define DBLN 200
  #define DBLNSZ (DBLN * sizeof( char ))
  char dbline[DBLN];

  // calculate filename of TL database
  tl_wasprintf( &dbfilename, L"%ls\\tlpkg\\texlive.tlpdb", instroot );

  // read start of database
  dbfile = _wfopen( dbfilename, L"r" );
  if ( !dbfile ) {
    release[0] = '\0';
    return;
  }
  free( dbfilename );
  for ( i=0; i<200; i++ ) {
    memset( dbline, 0, DBLNSZ );
    cptr = fgets( dbline, DBLN, dbfile );
    if( !cptr ) {
      // i.e. EOF or error
      fclose( dbfile );
      release[0] = '\0';
      return;
    }
    l = strlen( "depend release/" );
    if( !strncmp( "depend release/", dbline, l)) {
      // next 4 chars should be digits
      for( j=0; j<4; j++ ) {
        release[j] = dbline[l+j];
        if( release[j] < '0' || release[j] > '9' ) {
          memset( release, 0, 5 * sizeof( char ));
          break;
        }
      }
      release[4] = '\0';
      fclose( dbfile );
      return;
    }
  }
  // if we got here, release should still consist
  // of zeroes from the initial default initialization
  fclose( dbfile );
  return;
}

BOOL in_HKLM() {
  /*
  Check for uninstall reg entry in HKLM for current TL.

  On w64, registry entries under HKLM\software\wow6432node will be
  transparently aliased to corresponding entries under HKLM\software
  so for 32-bits programs so no worries on this score.

  For just checking the existence of the uninstall registry key,
  we can get away with ascii/ansi/utf8.
  */
  long result;
  HKEY uninst_key;
  char keyname[] =
     "Software\\Microsoft\\windows\\currentversion\\uninstall\\TeXLivexxxx";

  if (!release[0]) return FALSE;
  strncpy( keyname + strlen( keyname ) - 4, release, 4 );

  result = RegOpenKeyExA(
    HKEY_LOCAL_MACHINE,
    keyname,
    0,
    KEY_QUERY_VALUE,
    &uninst_key);
  // An additional check: compare the path of the uninstaller
  // with the path of this program, using
  // the GetFileInformationByHandle API call, this time using wchar_t
  return ( result == ERROR_SUCCESS );
}

BOOL is_elevated() {

  BOOL ret = FALSE;

  // http://stackoverflow.com/questions/8046097/
  HANDLE ProcessHandle, TokenHandle = NULL;
  TOKEN_ELEVATION Elevation;
  DWORD cbsize = sizeof( TOKEN_ELEVATION );
  ProcessHandle = GetCurrentProcess(); // no need to close afterwards
  if( OpenProcessToken( ProcessHandle, TOKEN_QUERY, &TokenHandle )) {
    if( GetTokenInformation(
        TokenHandle,
        TokenElevation, // NOT a variable!
        &Elevation,
        sizeof( Elevation ),
        &cbsize )) {
      ret = Elevation.TokenIsElevated;
      if( TokenHandle ) CloseHandle( TokenHandle );
    }
  }
  return ret;
}

// wrapper for GetEnvironmentVariable
wchar_t * get_env( wchar_t * ev ) {
  DWORD sz;
  wchar_t * val, dummy;
  sz = GetEnvironmentVariable( ev, &dummy, 1 );
  if ( sz<2 ) return NULL; // we do not want an empty string
  val = calloc_fatal( ++sz, sizeof( wchar_t ));
  sz = GetEnvironmentVariable( ev, val, sz );
  if ( !sz ) return NULL;
  val[sz] = L'\0';
  return val;
}


void add_to_searchpath( wchar_t * newdir ) {
  // newdir should be relative to instroot, without initial backslash
  wchar_t * oldpath, * newpath;
  oldpath = get_env( L"PATH" );
  if ( wcslen( oldpath ) > 0 ) {
    tl_wasprintf( &newpath, L"%ls\\%ls;%ls", instroot, newdir, oldpath );
    FREE0( oldpath );
    SetEnvironmentVariable( L"PATH", newpath );
  }
}

int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
    PWSTR cmdline, int iShow ) {
  DWORD l;
  wchar_t * wptr, * own_path, * wish, * tlshell, * tclargs;
  SHELLEXECUTEINFO shExecInfo;
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  int i;

  // calculate installation root: grandparent of own directory
  // also set texbin along the way
  // msdn is conflicted as to whether _wpgmptr is absolute;
  // don't use _get_wpgmptr
  own_path = calloc_fatal( 32768, sizeof( wchar_t ));
  l = GetModuleFileNameW( NULL, own_path, 32768 ); // length l excludes \0
  if ( l == 0 || GetLastError() != ERROR_SUCCESS )
    die( L"Tlshell: cannot get own path", 1 );

  tl_wasprintf( &instroot, L"%ls", own_path );
  // just copying, but let tl_wasprintf handle the allocation nuisance:
  for (i=0; i<3; i++) {
    wptr = wcsrchr( instroot, L'\\' );
    if ( !wptr ) die( L"Did not find TeX Live installation;\n"
      L"this is not the right tlshell.", 0 );
    *wptr = L'\0';
    if ( i==0 ) tl_wasprintf( &texbin, L"%ls", instroot );
  }

  // read TL release from tlpdb
  get_tl_release();
  if( ! release[0] ) {
    MessageBox( NULL, L"Problem with installation: cannot determine release!",
        L"Error", MB_OK | MB_ICONSTOP );
    return 1;
  }

  if ( in_HKLM( ) && ! is_elevated( )) {
    // this is the recommended way to get an elevation prompt(!)
    shExecInfo.cbSize = sizeof(SHELLEXECUTEINFO);
    shExecInfo.fMask = SEE_MASK_NO_CONSOLE;
    shExecInfo.hwnd = NULL;
    shExecInfo.lpVerb = L"runas";
    shExecInfo.lpFile = own_path;
    shExecInfo.lpParameters = cmdline;
    shExecInfo.lpDirectory = NULL;
    shExecInfo.nShow = SW_HIDE;
    shExecInfo.hInstApp = NULL;
    if ( !ShellExecuteEx( &shExecInfo ))
      die( L"Failed to restart elevated", 1 );
    return 0;
  }
  FREE0( own_path );

  add_to_searchpath( L"bin\\win32" );
  add_to_searchpath( L"tlpkg\\tlperl\\bin" );

  // locate wish/tclkit
  tl_wasprintf( &wish, L"%ls\\tlpkg\\tltcl\\tclkit.exe", instroot );
  if( FILE_EXISTS( wish )) {
    FREE0( wish );
    tl_wasprintf( &wish, L"tclkit.exe" );
    add_to_searchpath( L"tlpkg\\tltcl" );
  } else {
    tl_wasprintf( &wish, L"%ls\\tlpkg\\tltcl\\bin\\wish.exe", instroot );
    if( FILE_EXISTS( wish )) {
      FREE0( wish );
      tl_wasprintf( &wish, L"wish.exe" );
      add_to_searchpath( L"tlpkg\\tltcl\\bin" );
    } else {
      die( L"Cannot find tclkit or wish", 0 );
    }
  }
  // and tlshell.tcl, quoted if necessary
  tl_wasprintf( &tlshell,
      L"%ls\\texmf-dist\\scripts\\tlshell\\tlshell.tcl", instroot );
  if( ! FILE_EXISTS( tlshell )) die( L"tlshell not available", 0 );
  if ( wcschr( instroot, L' ' )) {
    FREE0( tlshell );
    tl_wasprintf( &tlshell,
        L"\"%ls\\texmf-dist\\scripts\\tlshell\\tlshell.tcl\"", instroot );
  }

  if ( cmdline && *cmdline )
    tl_wasprintf( &tclargs, L"%ls %ls -- %ls", wish, tlshell, cmdline );
  else
    tl_wasprintf( &tclargs, L"%ls %ls", wish, tlshell );

  // create wish/tclkit child process
  ZeroMemory( &si, sizeof(si) );
  si.cb = sizeof(si);
  si.dwFlags = STARTF_USESHOWWINDOW;
  si.wShowWindow = SW_HIDE ;
  ZeroMemory( &pi, sizeof(pi) );
  if( !CreateProcess(
    NULL,    // module name
    tclargs, // commandline, including program name
    NULL,    // process security atrributes
    NULL,    // thread security atrributes
    TRUE,    // handle inheritance
    0,       // creation flags
    NULL,    // pointer to environment block (uses parent if NULL)
    NULL,    // starting directory (uses parent if NULL)
    &si,     // STARTUPINFO structure
    &pi )    // PROCESS_INFORMATION structure
  ) die( L"Failed to start tlshell", 1 );
  return 0;
}
